# -*- coding: utf-8 -*-
"""
    tipfy.ext.auth.gaema
    ~~~~~~~~~~~~~~~~~~~~

    Authentication handler for `gaema`.

    :copyright: 2010 by tipfy.org.
    :license: BSD, see LICENSE.txt for more details.
"""
import datetime
import functools
import logging

from tipfy import local, HTTPException, RequestRedirect, Tipfy


class RequestAdapter(object):
    """Adapter to transform a :class:`tipfy.Request` object into a request
    with the attributes expected by `gaema`.

    It must define at least the following attributes or functions:

        request.arguments: a dictionary of GET parameters mapping to a list
                           of values.
        request.host: current request host.
        request.path: current request path.
        request.full_url(): a function returning the current full URL.
    """
    def __init__(self, request):
        """Initializes the request adapter.

        :param request:
            A :class:`tipfy.Request` instance.
        """
        self.arguments = dict(request.args)
        self.full_url = lambda: request.url
        self.host = request.host
        self.path = request.path


class AuthHandler(object):
    """Base authentication handler with common functions used by `gaema` mixin
    classes.
    """
    def __init__(self, request, **kwargs):
        """Depending on the auth mixins used, provide these configurations as
        keyword arguments:

        'google_consumer_key'
        'google_consumer_secret'

        'twitter_consumer_key'
        'twitter_consumer_secret'

        'friendfeed_consumer_key'
        'friendfeed_consumer_secret'

        'facebook_api_key'
        'facebook_secret'
        """
        self.request = RequestAdapter(request)
        self.settings = kwargs

    def require_setting(name, feature="this feature"):
        """Raises an exception if the given setting is not defined.

        :param name:
            Setting name.
        :param feature:
            Setting description.
        :return:
            `None`.
        """
        if not self.settings.get(name):
            raise Exception("You must define the '%s' setting in your "
                        "application to use %s" % (name, feature))

    def async_callback(self, callback, *args, **kwargs):
        """Wraps callbacks with this if they are used on asynchronous requests.

        Catches exceptions and properly finishes the request.

        :param callback:
            Callback function to be wrapped.
        :param args:
            Positional arguments fpr the callback.
        :param kwargs:
            Keyword arguments fpr the callback.
        :return:
            A wrapped callback.
        """
        if callback is None:
            return None

        if args or kwargs:
            callback = functools.partial(callback, *args, **kwargs)

        def wrapper(*args, **kwargs):
            try:
                return callback(*args, **kwargs)
            except Exception, e:
                logging.error('Exception during callback', exc_info=True)

        return wrapper

    def redirect(self, url):
        """Redirects to the given URL. We raise a `RequestRedirect`
        exception to be caught by the WSGI app.

        :param url:
            URL to redirect.
        :return:
            `None`.
        """
        raise RequestRedirect(url)

    _ARG_DEFAULT = []

    def get_argument(self, name, default=_ARG_DEFAULT, strip=True):
        """Returns the value of a request GET argument with the given name.

        If default is not provided, the argument is considered to be
        required, and we throw an HTTP 404 exception if it is missing.

        The returned value is always unicode.

        :param:
            Argument name to be retrieved from the request GET.
        :param default:
            Default value to be return if the argument is not in GET.
        :param strip:
            If `True`, returns the value after calling strip() on it.
        :return:
            A request argument, as unicode.
        """
        values = Tipfy.request.args.get(name, default)
        if value == _ARG_DEFAULT:
            raise HTTPException('Missing request argument %s' % name)

        if strip:
            value = value.strip()

        return value

    def get_cookie(self, name, default=None):
        """Gets the value of the cookie with the given name, else default.

        :param:
            Cookie name to be retrieved.
        :param default:
            Default value to be return if cookie is not set.
        :return:
            The cookie value.
        """
        return Tipfy.request.cookies.get(name, default)

    def set_cookie(self, name, value, domain=None, expires=None, path="/",
                   expires_days=None):
        """Sets the given cookie name/value with the given options.

        :param name:
            Cookie name.
        :param value:
            Cookie value.
        :param domain:
            Cookie domain.
        :param expires:
            A expiration date as a `datetime` object.
        :param path:
            Cookie path.
        :param expires_days:
            Number of days to calculate expiration.
        :return:
            `None`.
        """
        if expires_days is not None and not expires:
            expires = datetime.datetime.utcnow() + datetime.timedelta(
                days=expires_days)

        local.session_store.set_cookie(name, value, domain=domain, path=path,
            expires=expires)



# -*- coding: utf-8 -*-
"""
    tipfy.ext.auth.model
    ~~~~~~~~~~~~~~~~~~~~

    Base model for authenticated users.

    This module derives from `Zine`_.

    :copyright: 2010 by tipfy.org.
    :license: BSD, see LICENSE.txt for more details.
"""
import datetime
import string
from random import choice
from hashlib import sha1, md5

from google.appengine.ext import db

from tipfy import get_config

SALT_CHARS = string.ascii_letters + string.digits


class User(db.Model):
    """Universal user model. Can be used with App Engine's default users API,
    own auth or third party authentication methods (OpenId, OAuth etc).
    """
    #: Creation date.
    created = db.DateTimeProperty(auto_now_add=True)
    #: Modification date.
    updated = db.DateTimeProperty(auto_now=True)
    #: User defined unique name, also used as key_name.
    username = db.StringProperty(required=True)
    #: Password, only set for own authentication.
    password = db.StringProperty(required=False)
    #: User email
    email = db.EmailProperty()
    #: Authentication identifier, depending on the auth method used.
    #: own|username
    #: gae|user_id
    #: openid|identifier
    auth_id = db.StringProperty(required=True)
    # Session id, renewed periodically for improved security.
    session_id = db.StringProperty(required=True)
    # Session id last renewal date.
    session_updated = db.DateTimeProperty(auto_now_add=True)
    # Admin flag.
    is_admin = db.BooleanProperty(required=True, default=False)

    @classmethod
    def get_by_username(cls, username):
        return cls.get_by_key_name(username)

    @classmethod
    def get_by_auth_id(cls, auth_id):
        return cls.all().filter('auth_id =', auth_id).get()

    @classmethod
    def create(cls, username, auth_id, **kwargs):
        """Creates a new user and returns it. If the username already exists,
        returns None.

        :param username:
            Unique username.
        :param auth_id:
            Authentication id, according the the authentication method used.
        :param kwargs:
            Additional entity attributes.
        :return:
            The newly created user or ``None`` if the username already exists.
        """
        kwargs['username'] = username
        kwargs['key_name'] = username
        kwargs['auth_id'] = auth_id
        # Generate an initial session id.
        kwargs['session_id'] = gen_salt(length=32)

        if 'password_hash' in kwargs:
            # Password is already hashed.
            kwargs['password'] = kwargs.pop('password_hash')
        elif 'password' in kwargs:
            # Password is not hashed: generate a hash.
            kwargs['password'] = gen_pwhash(kwargs['password'])

        def txn():
            if cls.get_by_username(username) is not None:
                # Username already exists.
                return None

            user = cls(**kwargs)
            user.put()
            return user

        return db.run_in_transaction(txn)

    def set_password(self, new_password):
        """Sets a new, plain password.

        :param new_password:
            A plain, not yet hashed password.
        :return:
            ``None``.
        """
        self.password = gen_pwhash(new_password)

    def check_password(self, password):
        """Checks if a password is valid. This is done with form login

        :param password:
            Password to be checked.
        :return:
            ``True`` is the password is valid, ``False`` otherwise.
        """
        if check_password(self.password, password) != True:
            return False

        # Check if session id needs to be renewed.
        self._renew_session()
        return True

    def check_session(self, session_id):
        """Checks if a session id is valid.

        :param session_id:
            Session id to be checked.
        :return:
            ``True`` is the session id is valid, ``False`` otherwise.
        """
        if self.session_id != session_id:
            return False

        # Check if session id needs to be renewed.
        self._renew_session()
        return True

    def _renew_session(self, force=False):
        """Renews the session id if its expiration time has passed.

        :param force:
            ``True`` to force the session id to be renewed, ``False`` to check
            if the expiration time has passed.
        :return:
            ``None``.
        """
        if force is False:
            # Only renew the session id if it is too old.
            expires = datetime.timedelta(seconds=get_config('tipfy.ext.auth',
                'session_id_max_age'))

            force = (self.session_updated + expires < datetime.datetime.now())

        if force is True:
            self.session_id = gen_salt(length=32)
            self.session_updated = datetime.datetime.now()
            self.put()

    def __unicode__(self):
        """Returns this entity's username.

        :return:
            Username, as unicode.
        """
        return unicode(self.username)

    def __str__(self):
        """Returns this entity's username.

        :return:
            Username, as unicode.
        """
        return self.__unicode__()

    def __eq__(self, obj):
        """Compares this user entity with another one.

        :return:
            ``True`` if both entities have same key, ``False`` otherwise.
        """
        if not obj:
            return False

        return str(self.key()) == str(obj.key())

    def __ne__(self, obj):
        """Compares this user entity with another one.

        :return:
            ``True`` if both entities don't have same key, ``False`` otherwise.
        """
        return not self.__eq__(obj)


def gen_salt(length=10):
    """Generates a random string of SALT_CHARS with specified ``length``.

    :param length:
        Length of the salt.
    :return:
        A random salt.
    """
    if length <= 0:
        raise ValueError('requested salt of length <= 0')

    return ''.join(choice(SALT_CHARS) for i in xrange(length))


def gen_pwhash(password):
    """Returns the password encrypted in sha1 format with a random salt.

    :param password:
        Password to e hashed and formatted.
    :return:
        A hashed and formatted password.
    """
    if isinstance(password, unicode):
        password = password.encode('utf-8')

    salt = gen_salt()
    h = sha1()
    h.update(salt)
    h.update(password)
    return 'sha1$%s$%s' % (salt, h.hexdigest())


def check_password(pwhash, password):
    """Checks a password against a given hash value. Since  many systems save
    md5 passwords with no salt and it's technically impossible to convert this
    to a sha hash with a salt we use this to be able to check for legacy plain
    or salted md5 passwords as well as salted sha passwords::

        plain$$default

    md5 passwords without salt::

        md5$$c21f969b5f03d33d43e04f8f136e7682

    md5 passwords with salt::

        md5$123456$7faa731e3365037d264ae6c2e3c7697e

    sha passwords::

        sha1$123456$118083bd04c79ab51944a9ef863efcd9c048dd9a

    >>> check_password('plain$$default', 'default')
    True
    >>> check_password('sha1$$5baa61e4c9b93f3f0682250b6cf8331b7ee68fd8', 'password')
    True
    >>> check_password('sha1$$5baa61e4c9b93f3f0682250b6cf8331b7ee68fd8', 'wrong')
    False
    >>> check_password('md5$xyz$bcc27016b4fdceb2bd1b369d5dc46c3f', u'example')
    True
    >>> check_password('sha1$5baa61e4c9b93f3f0682250b6cf8331b7ee68fd8', 'password')
    False
    >>> check_password('md42$xyz$bcc27016b4fdceb2bd1b369d5dc46c3f', 'example')
    False

    :param pwhash:
        Hash to be checked.
    :param password:
        Password to be checked.
    :return:
        ``True`` if the password is valid, ``False`` otherwise.
    """
    if isinstance(password, unicode):
        password = password.encode('utf-8')

    if pwhash.count('$') < 2:
        return False

    method, salt, hashval = pwhash.split('$', 2)

    if method == 'plain':
        return hashval == password
    elif method == 'md5':
        h = md5()
    elif method == 'sha1':
        h = sha1()
    else:
        return False

    h.update(salt)
    h.update(password)
    return h.hexdigest() == hashval



# -*- coding: utf-8 -*-
"""
    tipfy.ext.auth.acl
    ~~~~~~~~~~~~~~~~~~

    Simple Access Control List.

    This module provides utilities to manage permissions for anything that
    requires some level of restriction, like datastore models or handlers.
    Access permissions can be grouped in roles for convenience, so that a new
    user can be assigned to a role directly instead of having all his
    permissions defined manually. Individual access permissions can then
    override or extend the role permissions.

    .. note::
       Roles are optional, so this module doesn't define a roles model, to keep
       things simple and fast. Role definitions are set directly in the Acl
       class. The strategy to load roles is open to the implementation: for
       best performance, define them statically in a module.

    Usage example:

    .. code-block:: python

       # Set a dict of roles with an 'admin' role that has full access and
       # assign users to it. Each role maps to a list of rules. Each rule is a
       # tuple (topic, name, flag), where flag is as bool to allow or disallow
       # access. Wildcard '*' can be used to match all topics and/or names.
       Acl.roles_map = {
           'admin': [
               ('*', '*', True),
           ],
       }

       # Assign users 'user_1' and 'user_2' to the 'admin' role.
       AclRules.insert_or_update(area='my_area', user='user_1',
           roles=['admin'])
       AclRules.insert_or_update(area='my_area', user='user_2',
           roles=['admin'])

       # Restrict 'user_2' from accessing a specific resource, adding a new
       # rule with flag set to False. Now this user has access to everything
       # except this resource.
       user_acl = AclRules.get_by_area_and_user('my_area', 'user_2')
       user_acl.rules.append(('UserAdmin', '*', False))
       user_acl.put()

       # Check that 'user_2' permissions are correct.
       acl = Acl(area='my_area', user='user_2')
       assert acl.has_access(topic='UserAdmin', name='save') is False
       assert acl.has_access(topic='AnythingElse', name='put') is True

    The Acl object should be created once after a user is loaded, so that
    it becomes available for the app to do all necessary permissions checkings.

    Based on concept from Solar's Access and Role classes: http://solarphp.com.

    :copyright: Paul M. Jones <pmjones@solarphp.com>
    :copyright: 2010 by tipfy.org.
    :license: BSD, see LICENSE.txt for more details.
"""
from google.appengine.ext import db
from google.appengine.api import memcache

from tipfy import cached_property, get_config
from tipfy.ext.db import PickleProperty

#: Cache for loaded rules.
_rules_map = {}


class AclMixin(object):
    """A mixin that adds an ``acl`` property to a
    :class:`tipfy.RequestHandler`.

    The handler must have the properties ``area`` and ``current_user`` set for
    it to work.
    """
    roles_map = None
    roles_lock = None

    @cached_property
    def acl(self):
        """Loads and returns the access permission for the currently logged in
        user. This requires the handler to have an ``area`` and
        ``current_user`` attributes. Casted to a string they must return the
        object identifiers.
        """
        return Acl(str(self.area.key()), str(self.current_user.key()),
            self.roles_map, self.roles_lock)


def validate_rules(rules):
    """Ensures that the list of rule tuples is set correctly."""
    assert isinstance(rules, list), 'Rules must be a list'

    for rule in rules:
        assert isinstance(rule, tuple), 'Each rule must be tuple'
        assert len(rule) == 3, 'Each rule must have three elements'
        assert isinstance(rule[0], basestring), 'Rule topic must be a string'
        assert isinstance(rule[1], basestring), 'Rule name must be a string'
        assert isinstance(rule[2], bool), 'Rule flag must be a bool'


class AclRules(db.Model):
    #: Creation date.
    created = db.DateTimeProperty(auto_now_add=True)
    #: Modification date.
    updated = db.DateTimeProperty(auto_now=True)
    #: Area to which this role is related.
    area = db.StringProperty(required=True)
    #: User identifier.
    user = db.StringProperty(required=True)
    #: List of role names.
    roles = db.StringListProperty()
    #: Lists of rules. Each rule is a tuple (topic, name, flag).
    rules = PickleProperty(validator=validate_rules)

    @classmethod
    def get_key_name(cls, area, user):
        """Returns this entity's key name, also used as memcache key.

        :param area:
            Area string identifier.
        :param user:
            User string identifier.
        :return:
            A key name.
        """
        return '%s:%s' % (str(area), str(user))

    @classmethod
    def get_by_area_and_user(cls, area, user):
        """Returns an AclRules entity for a given user in a given area.

        :param area:
            Area string identifier.
        :param user:
            User string identifier.
        :return:
            An AclRules entity.
        """
        return cls.get_by_key_name(cls.get_key_name(area, user))

    @classmethod
    def insert_or_update(cls, area, user, roles=None, rules=None):
        """Inserts or updates ACL rules and roles for a given user.

        :param area:
            Area string identifier.
        :param user:
            User string identifier.
        :param roles:
            List of the roles for the user.
        :param rules:
            List of the rules for the user.
        :return:
            An AclRules entity.
        """
        key_name = cls.get_key_name(area, user)

        def txn():
            user_acl = cls.get_by_key_name(key_name)
            if user_acl is None:
                user_acl = cls(key_name=key_name, area=area, user=user,
                    rules=[])

            if roles is not None:
                user_acl.roles = roles

            if rules is not None:
                user_acl.rules = rules

            user_acl.put()
            return user_acl

        return db.run_in_transaction(txn)

    @classmethod
    def get_roles_and_rules(cls, area, user, roles_map, roles_lock):
        """Returns a tuple (roles, rules) for a given user in a given area.

        :param area:
            Area string identifier.
        :param user:
            User string identifier.
        :param roles_map:
            Dictionary of available role names mapping to list of rules.
        :param roles_lock:
            Lock for the roles map: a unique identifier to track changes.
        :return:
            A tuple of (roles, rules) for the given user in the given area.
        """
        res = None
        cache_key = cls.get_key_name(area, user)
        if cache_key in _rules_map:
            res = _rules_map[cache_key]
        else:
            res = memcache.get(cache_key, namespace=cls.__name__)

        if res is not None:
            lock, roles, rules = res

        if res is None or lock != roles_lock or get_config('tipfy', 'dev'):
            entity = cls.get_by_key_name(cache_key)
            if entity is None:
                res = (roles_lock, [], [])
            else:
                rules = []
                # Apply role rules.
                for role in entity.roles:
                    rules.extend(roles_map.get(role, []))

                # Extend with rules, eventually overriding some role rules.
                rules.extend(entity.rules)

                # Reverse everything, as rules are checked from last to first.
                rules.reverse()

                # Set results for cache, applying current roles_lock.
                res = (roles_lock, entity.roles, rules)

            cls.set_cache(cache_key, res)

        return (res[1], res[2])

    @classmethod
    def set_cache(cls, cache_key, spec):
        """Sets a memcache value.

        :param cache_key:
            Cache key.
        :param spec:
            Value to be saved.
        """
        _rules_map[cache_key] = spec
        memcache.set(cache_key, spec, namespace=cls.__name__)

    @classmethod
    def delete_cache(cls, cache_key):
        """Deletes a memcache value.

        :param cache_key:
            Cache key.
        """
        if cache_key in _rules_map:
            del _rules_map[cache_key]

        memcache.delete(cache_key, namespace=cls.__name__)

    def put(self):
        """Saves the entity and clears the cache."""
        self.delete_cache(self.get_key_name(self.area, self.user))
        super(AclRules, self).put()

    def delete(self):
        """Deletes the entity and clears the cache."""
        self.delete_cache(self.get_key_name(self.area, self.user))
        super(AclRules, self).delete()

    def is_rule_set(self, topic, name, flag):
        """Checks if a given rule is set.

        :param topic:
            A rule topic, as a string.
        :param roles:
            A rule name, as a string.
        :param flag:
            A rule flag, a boolean.
        :return:
            ``True`` if the rule already exists, ``False`` otherwise.
        """
        for rule_topic, rule_name, rule_flag in self.rules:
            if rule_topic == topic and rule_name == name and rule_flag == flag:
                return True

        return False


class Acl(object):
    """Loads access rules and roles for a given user in a given area and
    provides a centralized interface to check permissions. Each ``Acl`` object
    checks the permissions for a single user. For example:

    .. code-block:: python

        from tipfy.ext.auth.acl import Acl

        # Build an Acl object for user 'John' in the 'code-reviews' area.
        acl = Acl('code-reviews', 'John')

        # Check if 'John' is 'admin' in the 'code-reviews' area.
        is_admin = acl.is_one('admin')

        # Check if 'John' can approve new reviews.
        can_edit = acl.has_access('EditReview', 'approve')
    """
    #: Dictionary of available role names mapping to list of rules.
    roles_map = {}

    #: Lock for role changes. This is needed because if role definitions change
    #: we must invalidate existing cache that applied the previous definitions.
    roles_lock = None

    def __init__(self, area, user, roles_map=None, roles_lock=None):
        """Loads access privileges and roles for a given user in a given area.

        :param area:
            An area identifier, as a string.
        :param user:
            A user identifier, as a string.
        :param roles_map:
            A dictionary of roles mapping to a list of rule tuples.
        :param roles_lock:
            Roles lock string to validate cache. If not set, uses the
            application version id.
        """
        if roles_map is not None:
            self.roles_map = roles_map

        if roles_lock is not None:
            self.roles_lock = roles_lock
        elif self.roles_lock is None:
            # Set roles_lock default.
            self.roles_lock = get_config('tipfy', 'version_id')

        if area and user:
            self._roles, self._rules = AclRules.get_roles_and_rules(area, user,
                self.roles_map, self.roles_lock)
        else:
            self.reset()

    def reset(self):
        """Resets the currently loaded access rules and user roles.

        :return:
            ``None``.
        """
        self._rules = []
        self._roles = []

    def is_one(self, role):
        """Check to see if a user is in a role group.

        :param role:
            A role name, as a string.
        :return:
            Boolean ``True`` if the user is in this role group; ``False``
            otherwise.
        """
        return role in self._roles

    def is_any(self, roles):
        """Check to see if a user is in any of the listed role groups.

        :param roles:
            An iterable of role names.
        :return:
            Boolean ``True`` if the user is in any of the role groups;
            ``False`` otherwise.
        """
        for role in roles:
            if role in self._roles:
                return True

        return False

    def is_all(self, roles):
        """Check to see if a user is in all of the listed role groups.

        :param roles:
            An iterable of role names.
        :return:
            Boolean ``True`` if the user is in all of the role groups;
            ``False`` otherwise.
        """
        for role in roles:
            if role not in self._roles:
                return False

        return True

    def has_any_access(self):
        """Checks if the user has any access or roles.

        :return:
            Boolean ``True`` if the user has any access rule or role set;
            ``False`` otherwise.
        """
        if self._rules or self._roles:
            return True

        return False

    def has_access(self, topic, name):
        """Checks if the user has access to a topic/name combination.

        :param topic:
            A rule topic, as a string.
        :param roles:
            A rule name, as a string.
        :return:
            Boolean ``True`` if the user has access to this rule; ``False``
            otherwise.
        """
        if topic == '*' or name == '*':
            raise ValueError("has_access() can't be called passing '*'")

        for rule_topic, rule_name, rule_flag in self._rules:
            if (rule_topic == topic or rule_topic == '*') and \
                (rule_name == name or rule_name == '*'):
                # Topic and name matched, so return the flag.
                return rule_flag

        # No match.
        return False



# -*- coding: utf-8 -*-
"""
    tipfy.ext.auth
    ~~~~~~~~~~~~~~

    User authentication and permissions extension.

    :copyright: 2010 by tipfy.org.
    :license: BSD, see LICENSE.txt for more details.
"""
from datetime import datetime, timedelta

from google.appengine.api import users

from tipfy import (cached_property, get_config, Forbidden, import_string,
    local, redirect, RequestRedirect, Tipfy, url_for)
from tipfy.ext import session

#: Default configuration values for this module. Keys are:
#:
#: - ``auth_system``: The default authentication class, as a string. Default
#:   is ``tipfy.ext.auth.AppEngineAuth`` (uses App Engine's built in users
#:   system to login). To use own authentication or authentication with
#:   OpenId, OAuth, Google Accounts, Twitter, FriendFeed or Facebook (one,
#:   all or a mix of these), set it to ``tipfy.ext.auth.MultiAuth``.
#:
#: - ``user_model``: A subclass of ``db.Model`` used for authenticated users,.
#:   as a string. Default is ``tipfy.ext.auth.model:User``.
#:
#: - ``cookie_key``: Name of the autentication cookie. Default is
#:   'tipfy.ext.auth'
#:
#: - ``cookie_session_expires``: Session expiration time in seconds. Limits the
#:   duration of the contents of a cookie, even if a session cookie exists.
#:   If ``None``, the contents lasts as long as the cookie is valid. Default is
#:   ``None``.
#:
#: - ``cookie_max_age``: Time in seconds before the authentication cookie
#:   data is invalidated. Both persistent and non-persitent authentications
#:   are affected by this setting - the difference is that non-persitent
#:   authentication only lasts for the current browser session. Default is 1
#:   week.
#:
#: - ``cookie_domain``: Domain of the cookie. To work accross subdomains the
#:   domain must be set to the main domain with a preceding dot, e.g., cookies
#:   set for `.mydomain.org` will work in `foo.mydomain.org` and
#:   `bar.mydomain.org`. Default is ``None``, which means that cookies will
#:   only work for the current subdomain.
#:
#: - ``cookie_path``: Path in which the authentication cookie is valid.
#:   Default is `/`.
#:
#: - ``cookie_secure``: Make the cookie only available via HTTPS.
#:
#: - ``cookie_httponly``: Disallow JavaScript to access the cookie.
#:
#: - ``session_id_max_age``: Interval in seconds before a user session id is
#:   renewed. Default is 1 week.
default_config = {
    'auth_system': 'tipfy.ext.auth.AppEngineAuth',
    'user_model': 'tipfy.ext.auth.model.User',
    'cookie_key': 'tipfy.user',
    'cookie_session_expires': None,
    'cookie_max_age': 86400 * 7,
    'cookie_domain':   None,
    'cookie_path':     '/',
    'cookie_secure':   None,
    'cookie_httponly': False,
    'session_id_max_age': 86400 * 7,
}

#: Configured authentication system instance, cached in the module.
_auth_system = None


class AuthMiddleware(object):
    """:class:`tipfy.RequestHandler` middleware that loads and persists a
    user.
    """
    def pre_dispatch(self, handler):
        """Executes before a :class:`tipfy.RequestHandler` is dispatched. If
        it returns a response object, it will stop the pre_dispatch middleware
        chain and won't run the requested handler method, using the returned
        response instead. However, post_dispatch hooks will still be executed.

        :param handler:
            A :class:`tipfy.RequestHandler` instance.
        :return:
            A :class:`tipfy.Response` instance or None.
        """
        # Start user session.
        get_auth_system().login_with_session()

    def pre_dispatch_handler(self):
        """Called if auth is used as a WSGIApplication middleware."""
        get_auth_system().login_with_session()
        return None


class BaseAuth(object):
    """Base authentication adapter."""
    #: Endpoint to the handler that creates a new user account.
    signup_endpoint = 'auth/signup'

    #: Endpoint to the handler that logs in the user.
    login_endpoint = 'auth/login'

    #: Endpoint to the handler that logs out the user.
    logout_endpoint = 'auth/logout'

    @cached_property
    def user_model(self):
        """Returns the configured user model.

        :return:
            A :class:`tipfy.ext.auth.model.User` class.
        """
        return import_string(get_config(__name__, 'user_model'))

    def login_with_session(self):
        """Authenticates the current user using sessions.

        :return:
            ``None``.
        """
        local.user = None

    def login_with_form(self, username, password, remember=False):
        """Authenticates the current user using data from a form.

        :param username:
            Username.
        :param password:
            Password.
        :param remember:
            True if authentication should be persisted even if user leaves the
            current session (the "remember me" feature).
        :return:
            ``True`` if login was succesfull, ``False`` otherwise.
        """
        local.user = None
        return False

    def login_with_external_data(self, data, remember=False):
        """Authenticates using data provided by an external service, such as
        OpenId, OAuth (Google Account, Twitter, Friendfeed) or Facebook.

        :return:
            ``None``.
        """
        local.user = None

    def logout(self):
        """Logs out the current user.

        :return:
            ``None``.
        """
        local.user = None

    def get_current_user(self):
        """Returns the currently logged in user entity or ``None``.

        :return:
            A :class:`User` entity, if the user for the current request is
            logged in, or ``None``.
        """
        return getattr(local, 'user', None)

    def is_current_user_admin(self):
        """Returns ``True`` if the current user is an admin.

        :return:
            ``True`` if the user for the current request is an admin, ``False``
            otherwise.
        """
        if getattr(local, 'user', None) is not None:
            return local.user.is_admin

        return False

    def create_signup_url(self, dest_url):
        """Returns the signup URL for this request and specified destination
        URL. By default returns the URL for the endpoint
        :attr:`signup_endpoint`.

        :param dest_url:
            String that is the desired final destination URL for the user once
            signup is complete.
        :return:
            A URL to perform signup.
        """
        return url_for(self.signup_endpoint, redirect=dest_url, full=True)

    def create_login_url(self, dest_url):
        """Returns the login URL for this request and specified destination
        URL. By default returns the URL for the endpoint
        :attr:`login_endpoint`.

        :param dest_url:
            String that is the desired final destination URL for the user once
            login is complete.
        :return:
            A URL to perform login.
        """
        return url_for(self.login_endpoint, redirect=dest_url, full=True)

    def create_logout_url(self, dest_url):
        """Returns the logout URL for this request and specified destination
        URL. By default returns the URL for the endpoint
        :attr:`logout_endpoint`.

        :param dest_url:
            String that is the desired final destination URL for the user once
            logout is complete.
        :return:
            A URL to perform logout.
        """
        return url_for(self.logout_endpoint, redirect=dest_url, full=True)

    def is_authenticated(self):
        """Returns ``True`` if the current user is logged in.

        :return:
            ``True`` if the user for the current request is authenticated,
            ``False`` otherwise.
        """
        # This is not really true but it is our best bet since 3rd party auth
        # is handled elsewhere.
        return (getattr(local, 'user', None) is not None)

    def create_user(self, username, auth_id, **kwargs):
        """Saves a new user in the datastore for the currently logged in user,
        and returns it. If the username already exists, returns ``None``.

        :param username:
            The unique username for this user.
        :param kwargs:
            Extra keyword arguments accepted by
            :class:`tipfy.ext.auth.model.User`.
        :return:
            The new :class:`tipfy.ext.auth.model.User` entity, or ``None`` if
            the username already exists.
        """
        return self.user_model.create(username, auth_id, **kwargs)


class MultiAuth(BaseAuth):
    """Authentication using own users or third party services such as OpenId,
    OAuth (Google, Twitter, FriendFeed) and Facebook.
    """
    @cached_property
    def cookie_key(self):
        return get_config(__name__, 'cookie_key')

    @cached_property
    def cookie_args(self):
        """Returns the configured arguments to set an auth cookie.

        :return:
            A dictionary of keyword arguments for a cookie.
        """
        keys = ['domain', 'path', 'secure', 'httponly', 'session_expires',
            'max_age']
        return dict((k, get_config(__name__, 'cookie_' + k)) for k in keys)

    def login_with_session(self):
        local.user = None

        # Check if a secure cookie is set.
        session = local.session_store.load_secure_cookie(self.cookie_key)

        if session.new is True:
            # Session didn't exist.
            return

        # Check if we are in the middle of external auth and account creation.
        if session.get('to_signup', None):
            # Redirect to account creation page.
            if not _is_auth_endpoint('signup_endpoint'):
                raise RequestRedirect(create_signup_url(Tipfy.request.url))

            return

        # Get the authentication and session ids.
        auth_id = session.get('id', None)
        session_id = session.get('session_id', None)
        remember = session.get('remember', '0')

        if auth_id is not None:
            user = self.user_model.get_by_auth_id(auth_id)
            if user is not None and user.check_session(session_id) is True:
                # Successful login. Make the user available.
                local.user = user
                # Add session to be saved at the end of request.
                # Reset session id in case it was renewed by the model.
                self.set_session(session, id=auth_id,
                    session_id=user.session_id, remember=remember)

    def login_with_form(self, username, password, remember=False):
        local.user = None
        user = self.user_model.get_by_username(username)

        if user is not None and user.check_password(password) is True:
            # Successful login. Make the user available.
            local.user = user
            # Store the cookie.
            self.set_session(id=user.auth_id, session_id=user.session_id,
                remember=str(int(remember)))
            return True

        # Authentication failed.
        return False

    def login_with_external_data(self, data, remember=False):
        local.user = None

        auth_id = data.get('auth_id', None)
        if auth_id is None:
            return

        user = self.user_model.get_by_auth_id(auth_id)

        if user is not None:
            local.user = user
            # Store the cookie.
            self.set_session(id=user.auth_id, session_id=user.session_id,
                remember=str(int(remember)))
        else:
            # Save temporary data while the user haven't created an account.
            self.set_session(to_signup=data)

            # Redirect to account creation page.
            # TODO: document this redirect.
            if not _is_auth_endpoint('signup_endpoint'):
                raise RequestRedirect(create_signup_url(Tipfy.request.url))

    def logout(self):
        local.user = None
        # Delete the cookie.
        local.session_store.set_cookie(self.cookie_key, None,
            **self.cookie_args)

    def set_session(self, session=None, **kwargs):
        """Renews the auth session."""
        if session:
            session.clear()
        else:
            session = local.session_store.create_secure_cookie()

        session.update(kwargs)

        # TODO add kwargs.
        cookie_args = dict(self.cookie_args)

        if kwargs.get('remember', '0') == '0':
            # Non-persistent authentication (only lasts for a session).
            cookie_args['max_age'] = None

        local.session_store.set_cookie(self.cookie_key, session, **cookie_args)

    def create_user(self, username, auth_id, **kwargs):
        user = self.user_model.create(username, auth_id, **kwargs)
        if user:
            self.set_session(id=user.auth_id, session_id=user.session_id,
                remember=0)

        return user


class AppEngineAuth(BaseAuth):
    """Authentication using App Engine's users module."""
    def login_with_session(self):
        local.user = None

        current_user = users.get_current_user()
        if current_user is None:
            return

        local.user = self.user_model.get_by_auth_id('gae|%s' %
            current_user.user_id())

        if local.user is None and not _is_auth_endpoint('signup_endpoint'):
            # User is logged in, but didn't create an account yet.
            raise RequestRedirect(create_signup_url(Tipfy.request.url))

    def create_login_url(self, dest_url):
        return users.create_login_url(dest_url)

    def create_logout_url(self, dest_url):
        return users.create_logout_url(dest_url)

    def is_authenticated(self):
        return (users.get_current_user() is not None)


def get_auth_system():
    """Returns the configured authentication system.

    :return:
        An instance of :class:`tipfy.ext.auth.BaseAuth`.
    """
    global _auth_system
    if _auth_system is None:
        _auth_system = import_string(get_config(__name__, 'auth_system'))()

    return _auth_system


def create_signup_url(dest_url):
    """Returns a URL that, when visited, prompts the user to sign up.

    Args:
        dest_url: str: A full URL or relative path to redirect to after signing
        up.
    Returns:
        str: A URL to perform sign up.
    """
    return get_auth_system().create_signup_url(dest_url)


def create_login_url(dest_url):
    """Returns a URL that, when visited, prompts the user to sign in.

    Args:
        dest_url: str: A full URL or relative path to redirect to after logging
        in.
    Returns:
        str: A URL to perform login.
    """
    return get_auth_system().create_login_url(dest_url)


def create_logout_url(dest_url):
    """Returns a URL that, when visited, logs out the user.

    Args:
        dest_url: str: A full URL or relative path to redirect to after logging
        out.
    Returns:
        str: A URL to perform logout.
    """
    return get_auth_system().create_logout_url(dest_url)


def get_current_user():
    """Returns a :class:`User` entity for the the currently logged in user or
    ``None``.

    :return:
        A :class:`User` entity if the user for the current request is logged
        in, or ``None``.
    """
    return get_auth_system().get_current_user()


def is_current_user_admin():
    """Returns ``True`` if the current user is an admin.

    :return:
        ``True`` if the user for the current request is an admin, ``False``
        otherwise.
    """
    return get_auth_system().is_current_user_admin()


def is_authenticated():
    """Returns ``True`` if the user is logged in. It is possible that a user
    is logged in and :func:`get_current_user()` is None if the user is
    authenticated but still haven't created an account.

    :return:
        ``True`` if the user for the current request is logged in, ``False``
        otherwise.
    """
    return get_auth_system().is_authenticated()


def login_required(func):
    """A handler decorator to require user authentication. Normally
    :func:`user_required` is used instead.

    :param func:
        The handler method to be decorated.
    :return:
        The decorated method.
    """
    def decorated(*args, **kwargs):
        if is_authenticated() or _is_auth_endpoint('login_endpoint'):
            return func(*args, **kwargs)

        # Redirect to login page.
        return redirect(create_login_url(Tipfy.request.url))

    return decorated


def user_required(func):
    """A handler decorator to require the current user to be authenticated and
    to have an account saved in datastore.

    :param func:
        The handler method to be decorated.
    :return:
        The decorated method.
    """
    def decorated(*args, **kwargs):
        if get_current_user() or _is_auth_endpoint(('signup_endpoint',
            'login_endpoint')):
            return func(*args, **kwargs)

        if is_authenticated():
            # Redirect to signup page.
            return redirect(create_signup_url(Tipfy.request.url))
        else:
            # Redirect to login page.
            return redirect(create_login_url(Tipfy.request.url))

    return decorated


def admin_required(func):
    """A handler decorator to require the current user to be admin.

    :param func:
        The handler method to be decorated.
    :return:
        The decorated method.
    """
    def decorated(*args, **kwargs):
        if is_current_user_admin() or _is_auth_endpoint('login_endpoint'):
            return func(*args, **kwargs)

        if not is_authenticated():
            # Redirect to signup page.
            return redirect(create_login_url(Tipfy.request.url))

        # Nope, user isn't an admin.
        raise Forbidden()

    return decorated


def basic_auth_required(validator):
    """A decorator to authenticate using HTTP basic/digest authorization.
    This only wraps a handler method by a validator function.

    :param validator:
        A validator function that accepts a
        `werkzeug.datastructures.Authorization` to validate the request, plus
        the wrapped handler method and arguments.
    :return:
        The decorated method.
    """
    def wrapper(func):
        def decorated(*args, **kwargs):
            return validator(Tipfy.request.authorization, func, *args,
                **kwargs)

        return decorated

    return wrapper


def _is_auth_endpoint(endpoints):
    """Helper function to check if the current url is the given auth endpoint.

    :endpoint:
        An endpoint name, as a string, or a tuple of endpoint names.
    :return:
        ``True`` if the current url is the given auth endpoint.
    """
    if isinstance(endpoints, basestring):
        endpoints = (endpoints,)

    auth_system = get_auth_system()
    rule = Tipfy.request.rule
    for e in endpoints:
        if rule.endpoint == getattr(auth_system, e, None):
            return True

    return False
